// 0xXYYZZ = X.YY.ZZ
const int MILKYTRACKER_VERSION = 0x10501;
// Version string restricted to 20 chars for XM export!
const char MILKYTRACKER_VERSION_STRING[] = "MilkyTracker 1.05.01";
//                                          012345678901234567890
