/*
 * Copyright (c) 2009, The MilkyTracker Team.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the <ORGANIZATION> nor the names of its contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  AudioDriverManager.h
 *
 *  This class serves as provider and manager of different kind of
 *	AudioDriver instances. Different kinds of AudioDrivers can be
 *	requested on the same system. 
 *
 *
 */
#ifndef __AUDIODRIVERMANAGER_H__
#define __AUDIODRIVERMANAGER_H__

#include "MilkyPlayTypes.h"

class AudioDriverInterface;

class AudioDriverManager
{
private:
	AudioDriverInterface**	driverList;
	mp_sint32				numDrivers;
	mp_sint32				defaultDriverIndex;
	mutable mp_sint32		enumerationIndex;
	
public:
	AudioDriverManager();
	~AudioDriverManager();

	/**
	 * Get the preferred audio driver for this system
	 * NULL is returned if either none is available or an alloc error occurred
	 * @return			audio driver instance
	 */
	AudioDriverInterface* getPreferredAudioDriver();
	AudioDriverInterface* getAudioDriverByName(const char* name);

	const char* getFirstDriverName() const;
	const char* getNextDriverName() const;
	
	mp_sint32 getPreferredAudioDriverSampleRate() const;
	mp_sint32 getPreferredAudioDriverBufferSize() const;
};

#endif
