/*
 * Copyright (c) 2009, The MilkyTracker Team.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the <ORGANIZATION> nor the names of its contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  AudioDriver_COMPENSATE.h
 *  MilkyPlay
 *
 *  Created by Peter Barth on 22.04.06.
 *
 */
#ifndef __AUDIODRIVER_COMPENSATE_H__
#define __AUDIODRIVER_COMPENSATE_H__

#include "AudioDriverBase.h"
#include "MilkyPlayCommon.h"
#include "MasterMixer.h"

class AudioDriver_COMPENSATE : public AudioDriverBase
{
protected:
	bool		deviceHasStarted;
	mp_uint32	sampleCounter;
	
public:
	AudioDriver_COMPENSATE() :
		deviceHasStarted(false),
		sampleCounter(0)
	{
	}

	virtual		~AudioDriver_COMPENSATE()
	{
	}

	virtual		mp_uint32	getNumPlayedSamples() const { return sampleCounter; }
	
	void fillAudioWithCompensation(char* stream, int length)
	{
		// sanity check
		if (!this->deviceHasStarted)
			return;
		
		MasterMixer* mixer = this->mixer;

		// Attention: Sample buffer MUST be 16 bit stereo, otherwise this will not work
		this->sampleCounter+=length>>2;
		//mixer->updateSampleCounter(length>>2);

		if (isMixerActive())
			mixer->mixerHandler((mp_sword*)stream);
		else
			memset(stream, 0, length);
	}
};

#endif

